@@ -102,7 +102,7 @@ MetaServerClient::~MetaServerClient() {
 
 
 void MetaServerClient::startAnnounce(const std::string& serverName, int serverPort, const std::string& mapName, Uint8 numPlayers, Uint8 maxPlayers,
-                                     const std::string& modName, const std::string& modVersion) {
+                                     const std::string& modName, const std::string& modVersion, uint16_t stunPort) {
 
     stopAnnounce();
 
@@ -114,8 +114,10 @@ void MetaServerClient::startAnnounce(const std::string& serverName, int serverPo
     this->maxPlayers = maxPlayers;
     this->modName = modName;
     this->modVersion = modVersion;
+    this->stunPort = stunPort;
+    this->sessionId = "";  // Will be set by response
 
-    enqueueMetaServerCommand(std::make_unique<MetaServerAdd>(serverName, serverPort, secret, mapName, numPlayers, maxPlayers, modName, modVersion));
+    enqueueMetaServerCommand(std::make_unique<MetaServerAdd>(serverName, serverPort, secret, mapName, numPlayers, maxPlayers, modName, modVersion, stunPort));
     lastAnnounceUpdate = SDL_GetTicks();
 }
 
@@ -158,6 +160,172 @@ void MetaServerClient::announceGameStart(const std::string& mapName, const std::
     enqueueMetaServerCommand(std::make_unique<MetaServerGameStart>(secret, mapName, modName, players, VERSIONSTRING));
 }
 
+// NAT Traversal / Hole Punch methods (synchronous)
+
+std::string MetaServerClient::requestHolePunch(const std::string& sessionId, uint16_t stunPort) {
+    std::map<std::string, std::string> parameters;
+    parameters["command"] = "punch_request";
+    parameters["session_id"] = sessionId;
+    parameters["stun_port"] = std::to_string(stunPort);
+    
+    try {
+        std::string result = loadFromHttp(metaServerURL, parameters);
+        
+        // Parse: OK\n<client_id>\n
+        std::istringstream stream(result);
+        std::string status;
+        std::getline(stream, status);
+        
+        // Trim CR
+        if (!status.empty() && status.back() == '\r') status.pop_back();
+        
+        if (status == "OK") {
+            std::string clientId;
+            std::getline(stream, clientId);
+            if (!clientId.empty() && clientId.back() == '\r') clientId.pop_back();
+            SDL_Log("MetaServerClient::requestHolePunch - got client_id: %s", clientId.c_str());
+            return clientId;
+        } else {
+            SDL_Log("MetaServerClient::requestHolePunch - failed: %s", result.c_str());
+            return "";
+        }
+    } catch (std::exception& e) {
+        SDL_Log("MetaServerClient::requestHolePunch - exception: %s", e.what());
+        return "";
+    }
+}
+
+bool MetaServerClient::pollPunchStatus(const std::string& sessionId, const std::string& clientId,
+                                       std::string& hostIP, uint16_t& hostPort, int& waitSeconds) {
+    std::map<std::string, std::string> parameters;
+    parameters["command"] = "punch_status";
+    parameters["session_id"] = sessionId;
+    parameters["client_id"] = clientId;
+    
+    try {
+        std::string result = loadFromHttp(metaServerURL, parameters);
+        
+        // Parse: WAITING\n or READY\n<host_ip>\n<host_port>\n<punch_in_seconds>\n
+        std::istringstream stream(result);
+        std::string status;
+        std::getline(stream, status);
+        
+        // Trim CR
+        if (!status.empty() && status.back() == '\r') status.pop_back();
+        
+        if (status == "READY") {
+            std::string line;
+            
+            std::getline(stream, line);
+            if (!line.empty() && line.back() == '\r') line.pop_back();
+            hostIP = line;
+            
+            std::getline(stream, line);
+            if (!line.empty() && line.back() == '\r') line.pop_back();
+            hostPort = static_cast<uint16_t>(std::stoi(line));
+            
+            std::getline(stream, line);
+            if (!line.empty() && line.back() == '\r') line.pop_back();
+            waitSeconds = std::stoi(line);
+            
+            SDL_Log("MetaServerClient::pollPunchStatus - READY: %s:%d in %d sec", 
+                    hostIP.c_str(), hostPort, waitSeconds);
+            return true;
+        } else if (status == "WAITING") {
+            return false;
+        } else {
+            SDL_Log("MetaServerClient::pollPunchStatus - unexpected: %s", result.c_str());
+            return false;
+        }
+    } catch (std::exception& e) {
+        SDL_Log("MetaServerClient::pollPunchStatus - exception: %s", e.what());
+        return false;
+    }
+}
+
+bool MetaServerClient::pollPunchRequests(std::vector<std::tuple<std::string, std::string, uint16_t>>& requests) {
+    if (secret.empty()) {
+        return false;  // Not hosting
+    }
+    
+    std::map<std::string, std::string> parameters;
+    parameters["command"] = "punch_poll";
+    parameters["secret"] = secret;
+    
+    try {
+        std::string result = loadFromHttp(metaServerURL, parameters);
+        
+        // Parse: OK\n[<client_id>\t<client_ip>\t<client_port>\n...]
+        std::istringstream stream(result);
+        std::string status;
+        std::getline(stream, status);
+        
+        // Trim CR
+        if (!status.empty() && status.back() == '\r') status.pop_back();
+        
+        if (status != "OK") {
+            SDL_Log("MetaServerClient::pollPunchRequests - failed: %s", result.c_str());
+            return false;
+        }
+        
+        requests.clear();
+        std::string line;
+        while (std::getline(stream, line)) {
+            if (!line.empty() && line.back() == '\r') line.pop_back();
+            if (line.empty()) continue;
+            
+            std::vector<std::string> parts = splitStringToStringVector(line, "\\t");
+            if (parts.size() >= 3) {
+                std::string clientId = parts[0];
+                std::string clientIP = parts[1];
+                uint16_t clientPort = static_cast<uint16_t>(std::stoi(parts[2]));
+                requests.push_back(std::make_tuple(clientId, clientIP, clientPort));
+                SDL_Log("MetaServerClient::pollPunchRequests - request from %s:%d (id: %s)", 
+                        clientIP.c_str(), clientPort, clientId.c_str());
+            }
+        }
+        
+        return true;
+    } catch (std::exception& e) {
+        SDL_Log("MetaServerClient::pollPunchRequests - exception: %s", e.what());
+        return false;
+    }
+}
+
+bool MetaServerClient::signalPunchReady(const std::string& clientId) {
+    if (secret.empty()) {
+        return false;  // Not hosting
+    }
+    
+    std::map<std::string, std::string> parameters;
+    parameters["command"] = "punch_ready";
+    parameters["secret"] = secret;
+    parameters["client_id"] = clientId;
+    
+    try {
+        std::string result = loadFromHttp(metaServerURL, parameters);
+        
+        // Parse: OK\n or ERROR\n<message>
+        std::istringstream stream(result);
+        std::string status;
+        std::getline(stream, status);
+        
+        // Trim CR
+        if (!status.empty() && status.back() == '\r') status.pop_back();
+        
+        if (status == "OK") {
+            SDL_Log("MetaServerClient::signalPunchReady - signaled ready for %s", clientId.c_str());
+            return true;
+        } else {
+            SDL_Log("MetaServerClient::signalPunchReady - failed: %s", result.c_str());
+            return false;
+        }
+    } catch (std::exception& e) {
+        SDL_Log("MetaServerClient::signalPunchReady - exception: %s", e.what());
+        return false;
+    }
+}
+
 
 void MetaServerClient::update() {
 
@@ -305,6 +473,12 @@ int MetaServerClient::connectionThreadMain(void* data) {
                     parameters["modname"] = pMetaServerAdd->modName;
                     parameters["modversion"] = pMetaServerAdd->modVersion;
                     
+                    // NAT traversal: Add STUN-discovered external port
+                    if (pMetaServerAdd->stunPort > 0) {
+                        parameters["stun_port"] = std::to_string(pMetaServerAdd->stunPort);
+                        SDL_Log("Announcing game with STUN port: %d", pMetaServerAdd->stunPort);
+                    }
+                    
                     // Add local IP for NAT traversal (allows clients on same LAN to connect directly)
                     std::string localIP = getLocalIPAddress();
                     if (!localIP.empty()) {
@@ -326,6 +500,27 @@ int MetaServerClient::connectionThreadMain(void* data) {
                         const std::string errorMsg = result.substr(result.find_first_not_of("\x0D\x0A",5), std::string::npos);
 
                         pMetaServerClient->setErrorMessage(METASERVERCOMMAND_ADD, errorMsg);
+                    } else {
+                        // Parse session_id from response (line 3)
+                        // Response format: OK\n<secret>\n<session_id>\n
+                        std::istringstream resultStream(result);
+                        std::string line;
+                        int lineNum = 0;
+                        while (std::getline(resultStream, line)) {
+                            lineNum++;
+                            // Trim CR if present
+                            if (!line.empty() && line.back() == '\r') {
+                                line.pop_back();
+                            }
+                            if (lineNum == 3 && !line.empty()) {
+                                // Store session_id (thread-safe via mutex)
+                                SDL_LockMutex(pMetaServerClient->sharedDataMutex);
+                                pMetaServerClient->sessionId = line;
+                                SDL_UnlockMutex(pMetaServerClient->sharedDataMutex);
+                                SDL_Log("MetaServerClient: Received session_id: %s", line.c_str());
+                                break;
+                            }
+                        }
                     }
 
 
@@ -414,7 +609,9 @@ int MetaServerClient::connectionThreadMain(void* data) {
                 case METASERVERCOMMAND_LIST: {
                     std::map<std::string, std::string> parameters;
 
-                    parameters["command"] = "list";
+                    // Use list2 for NAT traversal fields (session_id, stun_port)
+                    // Falls back to list if list2 not available
+                    parameters["command"] = "list2";
                     parameters["gameversion"] = VERSION;
 
                     std::string result;
@@ -422,8 +619,15 @@ int MetaServerClient::connectionThreadMain(void* data) {
                     try {
                         result = loadFromHttp(pMetaServerClient->metaServerURL, parameters);
                     } catch(std::exception& e) {
-                        pMetaServerClient->setErrorMessage(METASERVERCOMMAND_LIST, e.what());
-                        break;
+                        // Try fallback to list if list2 fails
+                        SDL_Log("MetaServerClient: list2 failed, trying list: %s", e.what());
+                        parameters["command"] = "list";
+                        try {
+                            result = loadFromHttp(pMetaServerClient->metaServerURL, parameters);
+                        } catch(std::exception& e2) {
+                            pMetaServerClient->setErrorMessage(METASERVERCOMMAND_LIST, e2.what());
+                            break;
+                        }
                     }
 
                     std::istringstream resultstream(result);
@@ -451,7 +655,8 @@ int MetaServerClient::connectionThreadMain(void* data) {
                             // - Intermediate: 10 fields (with localIP, no mod)
                             // - New: 11 fields (with localIP, modname, empty modversion - trailing field dropped by regex)
                             // - New: 12 fields (with localIP, modname, modversion)
-                            if(parts.size() < 9 || parts.size() > 12) {
+                            // - list2: 14 fields (with session_id, stun_port)
+                            if(parts.size() < 9 || parts.size() > 14) {
                                 break;
                             }
 
@@ -502,6 +707,20 @@ int MetaServerClient::connectionThreadMain(void* data) {
                                 gameServerInfo.modName = "vanilla";
                                 gameServerInfo.modVersion = "";
                             }
+                            
+                            // Parse NAT traversal fields if available (13th and 14th fields from list2)
+                            if(parts.size() >= 14) {
+                                gameServerInfo.sessionId = parts[12];
+                                int stunPortVal = 0;
+                                if(parseString(parts[13], stunPortVal) && stunPortVal > 0 && stunPortVal <= 65535) {
+                                    gameServerInfo.stunPort = static_cast<uint16_t>(stunPortVal);
+                                    gameServerInfo.holePunchAvailable = !gameServerInfo.sessionId.empty();
+                                }
+                            } else {
+                                gameServerInfo.sessionId = "";
+                                gameServerInfo.stunPort = 0;
+                                gameServerInfo.holePunchAvailable = false;
+                            }
 
                             if(resultstream.good() == false) {
                                 break;
